/*
 * Copyright (c) 2002 Toukan Ltd
 * All rights reserved
 *
 * Authors: Sergey Lyubka
 * Description:
 *   BSD - specific code for x86 processors.
 *   this file is a part of xtrace project
 *
 * $Id: bsd-i386.c,v 1.7 2002/05/15 14:40:57 drozd Exp $
 */

#include <machine/reg.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <sys/ptrace.h>
#include <sys/wait.h>

#include <stdio.h>
#include <signal.h>

#include "xtrace.h"
#include "bsd.h"

static unsigned char BPT[] = { 0xcc };

/*
 * Get call data by stack pointer
 */
int
xtr_get_call_data(struct opt *op, struct calldata *cdp)
{
	static vaddr_t args[MAXARGS], *argsp[MAXARGS];
	struct bsddata *dp = (struct bsddata *) op->data;
	static struct reg regs;
	vaddr_t esp;
	int i, scno;

	for (i = 0; i < sizeof(argsp)/sizeof(argsp[0]); i++)
		argsp[i] = args + i;

	if (dp->regsfd < 0)
		return (-1);

	if (lseek(dp->regsfd, 0L, SEEK_SET) == -1)
		perror("xtr_get_call_data: lseek()");

	if (read(dp->regsfd, &regs, sizeof(regs)) != sizeof(regs)) {
		perror("xtr_get_call_data: read(/proc/pid/regs)");
		return (-1);
	}

	esp = regs.r_esp + sizeof(int);
		
	cdp->retval = regs.r_eax;
	cdp->argv = argsp;
	cdp->ip = regs.r_eip;

	if (op->flags & XTR_INSYSENTER) {
		
		op->scno = regs.r_eax;
		
		/* handle indirect syscalls */
		if (op->scno == SYS___syscall
				|| op->scno == SYS_syscall) {

			(void) lseek(dp->procfd, esp, SEEK_SET);
			(void) read(dp->procfd, &scno, sizeof(int));

			if (op->scno == SYS_syscall)
				/* int - aligned */
				esp += sizeof(int);
			else
				/* quad - aligned */
				esp += sizeof(quad_t);

			op->scno = scno;

#ifdef __DEBUG__
			(void) fprintf(stderr, "*** xtr_get_call_data:"
				       " indirect syscall %d\n", op->scno);
#endif /* __DEBUG__ */
		}
	}


	/*
	 * in the case of backtrace or function call, limit the
	 * number of arguments
	 */
	if (!(op->flags & XTR_INSYSENTER) || (op->flags & XTR_BACKTRACING))
		argsp[4] = NULL;

	if (cdp->fp == 0U)
		cdp->fp = regs.r_ebp;

	if (!(op->flags & XTR_INSYSEXIT)) {
		
		/* read MAXARGS arguments from stack */
		if (op->flags & XTR_INSYSENTER)
			(void) lseek(dp->procfd, esp, SEEK_SET);
		else if (op->flags & XTR_INFENTER)
			(void) lseek(dp->procfd, esp + 4, SEEK_SET);
		else
			(void) lseek(dp->procfd, cdp->fp + 8, SEEK_SET);

		(void) read(dp->procfd, args, sizeof(args));

		(void) xtr_read_mem(op, cdp->fp, &cdp->ofp, sizeof(vaddr_t));
		(void) xtr_read_mem(op, cdp->fp +4, &cdp->rta, sizeof(vaddr_t));
	}

	return (0);
}

/*
 * Set the breakpoint on address. That's 0xCC, the INT3 instruction
 */
int
xtr_set_bpt(struct opt *op, struct symbol *symp)
{
	int ret = 0;

	if (symp->flags & SYM_BPT_SET) {
#ifdef __DEBUG__
		(void) fprintf(stderr, "*** bpt already set for %s\n",
		    symp->name);
#endif /* __DEBUG__ */
		--ret;
	} else {
		(void) xtr_read_mem(op, symp->va, &symp->instr, sizeof(BPT));
		(void) xtr_write_mem(op, symp->va, BPT, sizeof(BPT));
		symp->flags |= SYM_BPT_SET;
	}

	return (ret);
}

/*
 * Clear breakpoint
 */
int
xtr_clear_bpt(struct opt *op, struct symbol *symp)
{
	int ret = 0;

	if (!(symp->flags & SYM_BPT_SET)) {
#ifdef __DEBUG__
		(void) fprintf(stderr, "*** bpt not set for %s\n", symp->name);
#endif /* __DEBUG__ */
		--ret;
	} else {
		(void) xtr_write_mem(op, symp->va, &symp->instr, sizeof(BPT));
		symp->flags &= ~SYM_BPT_SET;
#ifdef __DEBUG__
	(void) fprintf(stderr, "*** clearing bpt for %s (%#x)\n", symp->name, symp->va);
#endif /* __DEBUG__ */
	}

	return (ret);
}

/*
 * Handle the signal child received
 */
void
handle_signal(struct opt *op)
{
	struct bsddata *dp = (struct bsddata *) op->data;
	struct symbol *symp;
	struct reg regs;
#ifdef __DEBUG__
	unsigned int instr;
#endif /* __DEBUG__ */

#ifdef __DEBUG__
	(void) fprintf(stderr, "*** handle_signal (sig %d)\n", op->signo);
#endif /* __DEBUG__ */

	CLEAR_FLAGS(op);

	if (op->signo != SIGTRAP) {
		op->flags |= XTR_SIGNAL;
		return;
	}

	op->flags |= XTR_INFENTER;

	/*
	 * fetch the registers to find out where the child is stopped
	 */
	if (lseek(dp->regsfd, 0, SEEK_SET) == -1) {
#ifdef __DEBUG__
		(void) fprintf(stderr, "*** handle_signal: lseek(regs) failed");
#endif /* __DEBUG__ */
		return;
	}
	if (read(dp->regsfd, &regs, sizeof(regs)) != sizeof(regs)) {
#ifdef __DEBUG__
		(void) fprintf(stderr, "*** handle_signal: read(regs) failed");
#endif /* __DEBUG__ */
		return;
	}
	
	/* EIP is known now, so find the appropriate symbol  */
	if ((symp = xtr_find_symbol(regs.r_eip, 0)) == NULL) {
#ifdef __DEBUG__
		(void) fprintf(stderr, "*** no symbol for "
			       "address %#x\n", regs.r_eip);
#endif /* __DEBUG__ */
		return;
	}
#ifdef __DEBUG__
	(void) xtr_read_mem(op, regs.r_eip - 1, &instr, sizeof(instr));
	(void) fprintf(stderr, "*** bpt hit %s (%#x, EIP %#x -> %#x)\n",
		       symp->name, symp->va, regs.r_eip, instr);
#endif /* __DEBUG__ */

	/*
	 * restore old instruction
	 */
	if (xtr_clear_bpt(op, (struct symbol *) symp) != 0)
		return;

	/*
	 * SIGTRAP has been sent _after_ INT3 instruction, so we have to
	 * move the instruction pointer back 1 byte after we've restored
	 * the old instruction
	 */
	--regs.r_eip;
	if (write(dp->regsfd, &regs, sizeof(regs)) != sizeof(regs))
		perror("xtr_get_call_data: write(regs)");

	/*
	 * set the single-step flag and perform single step.
	 * this control command will just set the TF (trap flag) in i386
	 * EFLAGS register, so SIGTRAP will occur after the execution of
	 * restored instruction. we will ignore SIGTRAP in the signal handler
	 * if single step flag set.
	 */
	op->flags |= XTR_STEPPING;
	ctl(dp->ctlfd, "step");

	/*
	 * at this point signal handler must receive SIGTRAP, ignore it and
	 * clear single-step flag. install breakpoint again, and continue
	 * the child
	 */
	xtr_set_bpt(op, symp);
}
