/*
 * Copyright (c) 2002 Toukan Ltd
 * All rights reserved
 *
 * Authors: Sergey Lyubka
 * Description:
 *   BSD - specific code
 *   this file is a part of xtrace project
 *
 * $Id: bsd.c,v 1.7 2002/05/15 13:51:14 drozd Exp $
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ptrace.h>
#include <sys/wait.h>
#include <sys/syscall.h>

#include <stdio.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <strings.h>
#include <errno.h>

#include "xtrace.h"
#include "bsd.h"
#include "elf.h"

static struct bsddata __data;		/* bsd - specific data */
static void fini(struct opt *);
static void chldhndlr(int signo);	/* CIGCHLD signal handler */

void
chldhndlr(int signo)
{
	struct bsddata *dp = &__data;
#ifdef __DEBUG__
	const char *reason;
#endif /* __DEBUG__ */
	int s;

	dp->op->flags |= XTR_INHANDLER;
	
	(void) wait(&s);

#ifdef __DEBUG__
	if (WIFSIGNALED(s))
		reason = "signalled";
	else if (WIFEXITED(s))
		reason = "exited";
	else if (WIFSTOPPED(s))
		reason = "stopped";
	else
		reason = "unknown";

	(void) fprintf(stderr, "*** chldhndlr: [%s] (sig %d)\n",
	    reason, WSTOPSIG(s));
#endif /* __DEBUG__ */

	dp->op->signo = WSTOPSIG(s);

	/*
	 * ignore SIGTRAP that occurs because of single step or after execve
	 */
	if (dp->op->signo == SIGTRAP &&
	    ((dp->op->flags & XTR_STEPPING) || (dp->op->scno == SYS_execve))) {
#ifdef __DEBUG__
		if (dp->op->flags & XTR_STEPPING)
			(void) fprintf(stderr, "*** chldhndlr: single step,"
			    " ignoring\n");
#endif /* __DEBUG__ */
		dp->op->flags &= ~XTR_STEPPING;
#ifdef __DEBUG__
		if (dp->op->scno == SYS_execve)
			(void) fprintf(stderr, "*** chldhndlr: SIGTRAP in"
			    " execve, ignoring\n");
#endif /* __DEBUG__ */
	} else {
		handle_signal(dp->op);
		xtr_dump_call(dp->op);
	}

	dp->op->flags &= ~XTR_INHANDLER;
	(void) ctl(dp->ctlfd, "run");
}


int
xtr_stop_app(struct opt *op)
{
	struct bsddata *dp = (struct bsddata *) op->data;
#ifdef __DEBUG__
	const char *reason;
#endif /* __DEBUG__ */

#ifdef __DEBUG__
	(void) fprintf(stderr, "*** xtr_stop_app: stopping app... ");
#endif /* __DEBUG__ */

	if (ioctl(dp->procfd, PIOCWAIT, &dp->status) < 0) {
		perror("ioctl(PIOCWAIT)"); return (-1);
	}

	CLEAR_FLAGS(op);
	switch (dp->status.why) {
		case S_SCE: 
#ifdef __DEBUG__
			reason = "sysenter";
#endif /* __DEBUG__ */
			op->flags |= XTR_INSYSENTER;
			break;
		case S_EXEC:
#ifdef __DEBUG__
			reason = "execve";
#endif /* __DEBUG__ */
			op->flags |= XTR_INEXEC;
			break;
		case S_SCX:
#ifdef __DEBUG__
			reason = "sysexit";
#endif /* __DEBUG__ */
			op->flags |= XTR_INSYSEXIT;
			break;
		case S_SIG:
#ifdef __DEBUG__
			reason = "signal";
#endif /* __DEBUG__ */
			    op->flags |= XTR_SIGNAL;
			    op->signo = dp->status.val;
			    break;
		default:
#ifdef __DEBUG__
			    reason = "unknown";
#endif /* __DEBUG__ */
			    op->flags |= XTR_OTHER;
			    break;
	}

#ifdef __DEBUG__
	(void) fprintf(stderr, "[%s]\n", reason);
#endif /* __DEBUG__ */
				
	if (dp->status.why == S_EXIT) {
		op->flags |= XTR_CHILDEXITED;
		fini((struct opt *) op);
		exit(0);
	}

	return (0);
}

int
xtr_resume_app(struct opt *op)
{
	struct bsddata *dp = (struct bsddata *) op->data;
	int val = 0;

#ifdef __DEBUG__
	(void) fprintf(stderr, "*** xtr_resume_app: resuming\n");
#endif /* __DEBUG__ */

	if (ioctl(dp->procfd, PIOCCONT, val) < 0) {
		perror("ioctl(PIOCCONT)"); return (-1);
	}

	return (0);
}

/*
 * Initialize tracer
 * Parameters:
 *    op - current options
 *    slist - comma separated list of syscalls to trace
 *    flist - comma separated list of functions to trace
 * Return:
 *    0 on success, < 0 on error
 */
int
xtr_init(struct opt *op, char *slist, char *flist)
{
	char fname[128];
	struct bsddata *bdp = &__data;
	struct procfs_status status;
	unsigned long flags;

	UNUSED(slist);
	UNUSED(flist);
	
	/* register cleanup and SIGCLD handler functions, setup data ptr */
	bdp->op = op;
	op->fini = fini;
	op->data = &__data;
	op->child = chldhndlr;

	/*
	 * open a ctl, regs, mem  files. Remember old event mask.
	 * set new event mask
	 */

	(void) snprintf(fname, sizeof(fname), "/proc/%d/mem", op->pid);
	if ((bdp->procfd = open(fname, O_RDWR)) < 0) {
		perror("xtr_init: open(/proc/pid/mem, O_RDWR"); exit(1);
	}

	flags = fcntl(bdp->procfd, F_GETFL);
	flags |= FD_CLOEXEC;
	(void) fcntl(bdp->procfd, F_SETFL, flags);

	(void) snprintf(fname, sizeof(fname), "/proc/%d/regs", op->pid);
	if ((bdp->regsfd = open(fname, O_RDWR)) < 0) {
		perror("xtr_init: open(/proc/pid/regs, O_RDWR"); exit(1);
	}
	
	(void) snprintf(fname, sizeof(fname), "/proc/%d/ctl", op->pid);
	if ((bdp->ctlfd = open(fname, O_WRONLY)) < 0) {
		perror("xtr_init: open(ctl)"); exit(1);
	}

	if (ioctl(bdp->procfd, PIOCSTATUS, &status) == -1) {
		perror("xtr_init: ioctl(PIOCSTATUS)"); exit(1);
	}
	bdp->oevnts = status.events;

	/* clear PF_LINGER flag */
	if (ioctl(bdp->procfd, PIOCGFL, &flags) == -1) {
		perror("xtr_init: ioctl(PIOCGFL)"); exit(1);
	}
	flags &= ~PF_LINGER;
	if (ioctl(bdp->procfd, PIOCSFL, flags) == -1) {
		perror("xtr_init: ioctl(PIOCSFL)"); exit(1);
	}
	
	bdp->evnts = S_SCE | S_SCX | S_EXIT | S_SIG | S_EXEC;
	if (ioctl(bdp->procfd, PIOCBIS, bdp->evnts) == -1) {
		perror("xtr_init: ioctl(PIOCBIS)"); exit(1);
	}

	(void) ctl(bdp->ctlfd, "attach");
	
	return (0);
}

/*
 * Dump memory content from process's address space
 */
int
xtr_read_mem(const struct opt *op, vaddr_t va, void *buf, int len)
{
	int ret;
	struct bsddata *dp = (struct bsddata *) op->data;

	(void) lseek(dp->procfd, va, SEEK_SET);
	
	if ((ret = read(dp->procfd, buf, len) != len)) {
#ifdef __DEBUG__
		perror("*** xtr_read_mem: read(va)");
#endif /* __DEBUG__ */
	}

	return (ret);
}

/*
 * Write into process's address space
 */
int
xtr_write_mem(const struct opt *op, vaddr_t va, void *buf, int len)
{
	int ret;
	struct bsddata *dp = (struct bsddata *) op->data;

	(void) lseek(dp->procfd, va, SEEK_SET);
	
	if ((ret = write(dp->procfd, buf, len) != len)) {
#ifdef __DEBUG__
		perror("*** xtr_write_mem: write(va)");
#endif /* __DEBUG__ */
	}

	return (ret);
}

void
xtr_before_exec(struct opt *op)
{
}

/*
 * Cleanup function. Must handle Linux-specific stuff.
 * See what xtr_fini() does before patching this code
 */
void
fini(struct opt *op)
{
	struct bsddata *bdp = (struct bsddata *) op->data;
		
	if (!(op->flags & XTR_CHILDEXITED)) {
		if (ioctl(bdp->procfd, PIOCBIS, bdp->oevnts) == -1) {
#ifdef __DEBUG__
			(void) fprintf(stderr, "*** fini: ioctl(PIOCBIS)\n");
#endif /* __DEBUG__ */
		}
	}

	(void) ctl(bdp->ctlfd, "detach");

	(void) close(bdp->procfd);
	(void) close(bdp->regsfd);
	(void) close(bdp->ctlfd);
}

/*
 * Send specified command `cmd' to a /proc/pid/ctl procfs control file
 */
int
ctl(int fd, const char *cmd)
{
	int ret = 0;
	int len = strlen(cmd);

	if (write(fd, cmd, len) != len) {
#ifdef __DEBUG__
		(void) fprintf(stderr, "*** ctl: %s (%s)\n",
		    cmd, strerror(errno));
#endif /* __DEBUG__ */
		--ret;
	}

	return (ret);
}
