/*
 * Copyright (c) 2002 Toukan Ltd
 * All rights reserved
 *
 * Authors: Sergey Lyubka
 * Description:
 *   dump call routines
 *   this file is a part of xtrace project
 *
 * $Id: dump.c,v 1.6 2002/05/15 13:51:14 drozd Exp $
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/syscall.h>
#include <sys/time.h>

#include <netinet/in.h>

#include <stdio.h>
#include <string.h>
#include <stdarg.h>

#include "xtrace.h"

static void dumparg(struct opt *op, const struct Arg *ap, vaddr_t va);
static void backtrace(struct opt *, struct calldata *);
static int tprint(const char *fmt, ...);
static void dumptime(void);

/*
 * Print out a time
 * Parameters, Return: void
 */
void
dumptime(void)
{
	struct timeval tv = { 0U, 0U };

	(void) gettimeofday(&tv, 0);
	(void) tprint("(%lu.%lu)", tv.tv_sec, tv.tv_usec);
}

/*
 * Print the sockaddr structure content pointed by va in debuggee's
 * address space.
 */
static void
dump_sockaddr(struct opt *op, vaddr_t va)
{
	struct sockaddr_in *sin;
	struct sockaddr sa;
	unsigned char *ucp;

	(void) xtr_read_mem(op, va, (void *)&sa, sizeof(sa));
	
	if (sa.sa_family == PF_INET) {
		sin = (struct sockaddr_in *) &sa;
		ucp = (unsigned char *) &sin->sin_addr;
		
		op->ncols += tprint("{family: PF_INET "
		    "port: %d addr: %d.%d.%d.%d}",
		    ntohs(sin->sin_port), *ucp, *(ucp + 1),
		    *(ucp + 2), *(ucp + 3));
	} else {
		op->ncols += tprint( "{ family: %d, data: %s }",
		    sa.sa_family, sa.sa_data);
	}
}

/*
 * Print the stat structure content pointed by va in debuggee's address
 * space. op->ncols incremented by the number of characters printed.
 */
static void
dump_stat(struct opt *op, vaddr_t va)
{
	struct stat st;
	
	(void) xtr_read_mem(op, va, (void *)&st, sizeof(st));
	
	op->ncols += tprint("dev: %d ino: %d mod: %d nlink: %d uid: %d gid: %d"
	    " atime: %lu ctime: %lu mtime: %lu size: %ld",
	    st.st_dev, st.st_ino, st.st_mode, st.st_nlink,
	    st.st_uid, st.st_gid, st.st_atime,
	    st.st_ctime, st.st_mtime, st.st_size);
}

/*
 * Print the fdset_t variable content pointed by va in debuggee's address
 * space. op->ncols incremented by the number of characters printed.
 */
static void
dump_fdset(struct opt *op, vaddr_t va)
{
	unsigned long ul;
	size_t siz;
	int first = 1;

	if (va == 0) {
		op->ncols += tprint("NULL");
		return;
	}

	(void) xtr_read_mem(op, va, (char *) &ul, sizeof(ul));
	op->ncols += tprint("[");
	
	for (siz = 0U; siz < 8 * sizeof(ul); siz++)
		if (ul & (1<<siz)) {
			if (first) {
				first = 0;
			} else {
				op->ncols += tprint(" ");
			}
			
			op->ncols += tprint("%u", siz);
		}
	op->ncols += tprint("]");
}

/*
 * Print the struct timeval content pointed by va in debuggee's address
 * space. op->ncols incremented by the number of characters printed.
 */
static void
dump_timeval(struct opt *op, vaddr_t va)
{
	struct timeval tv;

	if (va == 0) {
		op->ncols += tprint("NULL");
		return;
	}

	(void) xtr_read_mem(op, va, &tv, sizeof(tv));
	op->ncols += tprint("[%ld s, %ld ms]", tv.tv_sec, tv.tv_usec);
}

/*
 * Print the string content pointed by va in debuggee's address
 * space. op->ncols incremented by the number of characters printed.
 */
static void
dump_string(struct opt *op, vaddr_t va)
{
	char obuf[256];		/* we can jump out of readable memory */

	if (va == 0) {
		op->ncols += tprint("NULL");
		return;
	}

	(void) xtr_read_mem(op, va, obuf, sizeof(obuf));
	obuf[sizeof(obuf) - 1] = '\0';
	op->ncols += tprint("\"%s\"", obuf);
}

/*
 * Dump the argument pointed by va, depending of argument's type
 */
void
dumparg(struct opt *op, const struct Arg *ap, vaddr_t va)
{
	enum ArgType type;

	type = ap->type;
	
	if (ap->inout == OUT)
		type = HEX;

	switch (type) {
		case STAT:	dump_stat(op, va);	break;
		case SOCKADDR:	dump_sockaddr(op, va);	break;
		case FDSET:	dump_fdset(op, va);	break;
		case TIMESPEC:
				/* FALLTHROUGH */
		case TIMEVAL:	dump_timeval(op,va);	break;
		case INT: op->ncols += tprint("%d", va);break;
		case STRING:	dump_string(op, va);	break;
		
				/* FIXME */
		case MSGHDR:
		case SIGALTSTACK:
		case IOCTL:
		case ARGV:
		case ITIMER:
		case FCNTL:
		case TIMEZONE:
		case IOVEC:
		case STATFS:
		case RTPRIO:
		case RLIMIT:
		case MODSTAT:
		case KLDSTAT:
		case SHEDPARAM:
		case SIGSET:
		case SIGACTION:
		case KEVENT:
		case HEX: op->ncols += tprint("%#x", va);break;
		default: break;
	}
}

/*
* Dump information about current call
* Parameters:
*   sp - ptr to struct prstatus, should be read from /proc/<pid>/status
* Return:
*   void
 */
	void
xtr_dump_call(struct opt *op)
{
	const struct syscall *scp;
	struct calldata cd;
	ptrdiff_t diff;
	const struct symbol *symp;
	int i, first_arg, len;
	char padding[PADDING];
	static int oflags;		/* xtrace flags at previous dump */

	if ((op->flags & XTR_SIGNAL) && !(op->flags & XTR_INHANDLER)) {
#ifdef __DEBUG__
		(void) fprintf(stderr, "*** xtr_dump_call: signal %d,"
		    " but ignored (not in handler)\n", op->signo);
#endif /* __DEBUG__ */
		return;
	}

	/* get the data */
	(void) memset(&cd, 0, sizeof(struct calldata));
	if (xtr_get_call_data(op, &cd) != 0) {
#ifdef __DEBUG__
		(void) fprintf(stderr, "*** xtr_dump_call: call data unavailabe\n");
#endif /* __DEBUG__ */
		exit(1);
	}

	if (op->flags & XTR_INSYSEXIT) {

		/* clear interrupt flag if any */
		if (op->flags & XTR_INTERRUPTED) {
			op->flags &= ~XTR_INTERRUPTED;
			op->ncols = tprint("... continued. return value ");
		}

		/* print padding */
		len = op->ncols >= PADDING ? 0 : PADDING - op->ncols;
		(void) memset(padding, ' ', len);
		padding[len] = '\0';
		(void) tprint(padding);

		/* print syscall's return value */
		(void) tprint(" = %u ", cd.retval);

		/* print time stamp, if requested */
		if (op->flags & XTR_TIMESTAMP)
			dumptime();

		/* newline */
		(void) tprint("\n");

		/* rebuild symbol table after execve() */
		if (op->scno == SYS_execve) {
			xtr_symtab_destroy();
			xtr_symtab_build(op);
			if (op->flags & XTR_TRACEFUNC)
				xtr_set_func_breakpoints(op);
		}

		/* dump backtrace if requested */
		if (op->flags & XTR_BACKTRACE)
			backtrace(op, 0);

	} else if (op->flags & XTR_INSYSENTER) {

		op->ncols = 0;

		scp = xtr_lookup_syscall(op->scno);
		op->ncols += tprint("%s(", scp->name);

		/* print out the syscall's arguments */
		for (i = 0; i < scp->nargs; i++) {
			dumparg(op, &scp->args[i], *cd.argv[i]);
			if (i < scp->nargs -1)
				op->ncols += tprint(",");
		}

		op->ncols += tprint(")");
		if (op->scno == 1)
			tprint("\n");
	} else if (op->flags & XTR_INFENTER) {
		
		/* print the function name */
		symp = xtr_find_symbol(cd.ip, &diff);
		op->ncols = tprint("[%s](", symp->name);
		
		/* print function's parameters */
		for (i = 0, first_arg = 1; cd.argv[i] != NULL; i++) {
			if (first_arg)
				first_arg = 0;
			else
				op->ncols += tprint(",");
			
			op->ncols += tprint("%#x", *cd.argv[i]);
		}
		op->ncols += tprint(")\n");
	} else if (op->flags & XTR_SIGNAL) {
		if (oflags & XTR_INSYSENTER) {
			op->ncols = tprint(" -- interrupted\n");
			op->flags |= XTR_INTERRUPTED;
		}
		op->ncols = tprint("SIGNAL %d\n", op->signo);
	} else if (op->flags & XTR_INEXEC) {
		op->ncols = tprint("\nEXECVE\n");
	} else {
		op->ncols = tprint("unknown reason to stop\n");
	}

	/* remember  */
	oflags = op->flags;
}

/*
 * Dump call stack
 * Parameters:
 *    op - ptr to options
 * Return: void
 */
	void
backtrace(struct opt *op, struct calldata *cdp)
{
	ptrdiff_t diff;		/* offset from the function prologue */
	const struct symbol *symp;
	struct calldata cd;
	int leaf = 0;
	int first = 1;
	vaddr_t ip;
	static vaddr_t rta;	/* return address */

	op->flags |= XTR_BACKTRACING;

	/* set the frame pointer to 0 */
	if (cdp == 0) {
		cd.fp = (vaddr_t) 0;
		cdp = &cd;
		++leaf;
	}

	if (xtr_get_call_data(op, cdp) != 0)
		return;

	if (leaf == 1) {
		rta = cdp->rta;
		ip = cdp->ip;
	} else {
		ip = rta;
		rta = cdp->rta;
	}
	
	symp = xtr_find_symbol(ip, &diff);
	if (!symp)
		return;

	/* dump call data */
	(void) tprint("\t%-8x [%s + %x](", ip, symp->name, diff);
	for (; *cdp->argv != 0; cdp->argv++) {

		/* print the comma */
		if (first)
			--first;
		else
			(void) putchar(',');
		
		/* print parameter's value */
		(void) tprint("%x", **cdp->argv);
		
		/* optionally, dump memory */
		if (op->flags & XTR_DUMPMEM) {
			(void) xtr_read_mem(op, **cdp->argv, NULL, 0x10);
			/* FIXME: dump content onto stdout */
		}
	}
	(void) tprint(")\n", diff);

	op->flags &= ~XTR_BACKTRACING;
	
	if (cdp->ofp != 0U) {

		/* do that recursively  - shift one frame up */
		cdp->fp = cdp->ofp;
		backtrace(op, cdp);
	}
}

/*
* dump onto a terminal
 */
	int
tprint(const char *fmt, ...)
{
	va_list ap;
	int ret;

	va_start(ap, fmt);
	ret = vfprintf(stdout, fmt, ap);
	va_end(ap);

	(void) fflush(stdout);

	return (ret);
}
