/*
 * Copyright (c) 2002 Toukan Ltd
 * All rights reserved
 *
 * Authors: Sergey Lyubka
 * Description:
 *   Main module for xtrace utility
 *   this file is a part of xtrace project
 *
 * $Id: xtrace.c,v 1.11 2002/05/15 13:51:14 drozd Exp $
 */

#include <sys/types.h>
#include <sys/time.h>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>

#include "xtrace.h"

static const char __version[] =
"$Id: xtrace.c,v 1.11 2002/05/15 13:51:14 drozd Exp $";

/* options */
static struct opt __opt;

static void prepare(struct opt *, char *, char *);
static void start_app(struct opt *, char *const *, char *const *);
static void fini(int);

/*
 * Print usage information and exit
 * Parameters:
 *   pname - executable's name
 * Return value:
 *   void
 */
void usage(const char *pname)
{
	(void) fprintf(stderr, "usage:\n%s [OPTIONS] [filename ...]\n"
	    "OPTIONS:\n"
	    "-p <pid>\tPID of a process to be traced\n"
	    "-h\t\tthis help\n"
	    "-V\t\tprint version and exit\n"
	    "-S\t\tdump stack trace\n"
	    "-T\t\tdump per-thread info\n"
	    "-o <file>\tredirect output to file\n"
	    "-t\t\tput timestamp for every syscall\n"
	    "-v\t\tshow memory content for valid addresses\n"
	    "-s <syscall[,syscall2...]>\ttrace specific syscalls\n"
	    "-f <func1,[func2...]|all>\ttrace functions\n"
	    "-L\t\tdump list of functions from executable and exit\n"
	    "Note: -v, -S, -f options affect perfomance seriously!\n",
	    pname);
	exit(1);
}

/*
 * jump point for crt1.o code
 */
int
main(int argc, char *const *argv, char *const *envp)
{
	volatile int ch, retval;
	extern char *optarg;
	extern int optind;
	char *syslist = 0, *funclist = 0;

	while ((ch = getopt(argc, argv, "vVtTLSs:f:p:o:")) != -1)
		switch (ch) {
			case 'p':
				__opt.pid = (pid_t) atoi(optarg);
				break;
			case 't':
				__opt.flags |= XTR_TIMESTAMP;
				break;
			case 's':
				__opt.flags |= XTR_TRACESYS;
				syslist = optarg;
				break;
			case 'o':
				(void) freopen(optarg, "w", stdout);
				break;
			case 'v':
				__opt.flags |= XTR_DUMPMEM;
				break;
			case 'f':
				__opt.flags |= XTR_TRACEFUNC ;
				funclist = optarg;
				break;
			case 'S':
				__opt.flags |= XTR_BACKTRACE;
				break;
			case 'T':
				__opt.flags |= XTR_DUMPTHREADS;
				break;
			case 'L':
				__opt.flags |= XTR_FUNCLIST;
				break;
			case 'V':
				(void) printf("%s\n", __version);
				_exit(0);
				break;
			case 'h':
				/* FALLTHROUGH */
			default:
				usage(*argv);
				/* NOTREACHED */
				break;
		}

	if (argc == 1)
		usage(argv[0]);

	/* check child's PID sanity and start the child if requested */
	if (argc != optind) {
		start_app(&__opt, &argv[optind], envp);
	} else if (__opt.pid <= (pid_t) 0) {
		(void) fprintf(stderr, "invalid pid %d\n", (int)__opt.pid);
		exit(1);
	}

	prepare(&__opt, syslist, funclist);
	
	/* infinite trace loop */
	for (retval = -1; retval != 0; ) {

		/* stop on event of interest */
		(void) xtr_stop_app(&__opt);

		/* show it */
		xtr_dump_call(&__opt);

		/* resume process */
		(void) xtr_resume_app(&__opt);
	}

	return 0;
}

/*
 * Signal handler, and atexit handler in one
 * Parameters:
 *   signo - signal number caught
 * Return:
 *   void
 */
void
fini(int signo)
{
	struct opt *op = &__opt;
	struct symbol *symp;

	/* unset the breakpoints */
	for (symp = 0; (symp = xtr_get_next_sym()); )
		if (symp->flags & SYM_BPT_SET)
			xtr_clear_bpt(op, symp);
	
	/* call platform specific cleanup */
	if (__opt.fini != 0)
		(*__opt.fini)(&__opt);


	xtr_symtab_destroy();

	(void) fflush(stderr);
	(void) fflush(stdout);

	_exit(0);
}

/*
 * Platform-independet xtr_init()
 * Parameters:
 *    op - ptr to options
 *    slist - comma separated list of syscalls to trace
 *    flist - comma separated list of functions to trace
 * Return: void
 */
void
prepare(struct opt *op, char *slist, char *flist)
{
	/* trace syscalls by default */
	if (!(op->flags & XTR_TRACESYS) && !(op->flags & XTR_TRACEFUNC)) {
		op->flags |= XTR_TRACESYS;
		slist = "all";
	}

	op->slist = slist;
	op->flist = flist;

	(void) atexit((void (*)(void)) fini);

	(void) signal(SIGINT, fini);
	(void) signal(SIGTERM, fini);
	(void) signal(SIGQUIT, fini);

	/* call platform dependent init function */
	if (xtr_init(op, slist, flist) < 0) {
		(void) fprintf(stderr, "error init xtrace\n");
		exit(1);
	}

	/* set SIGCLD handler to fini if xtr_init() didn't set it */
	if (op->child != 0) {
		(void) signal(SIGCHLD, op->child);
	}

	/*
	 * FIXME: I must find a way to signal to parent when child has
	 * execve() completed so to rebuild symbol table
	 */

	/* fill up symbol table */
	if ((op->flags & XTR_TRACEFUNC) || (op->flags & XTR_BACKTRACE)
	    || (op->flags & XTR_FUNCLIST))
		(void) xtr_symtab_build(op);

	if (op->flags & XTR_TRACEFUNC)
		xtr_set_func_breakpoints(op);

	/* mark as attached */
	op->flags |= XTR_ATTACHED;
}

/*
 * Set the breakpoints on function prologues
 * Parameters:
 *   op - ptr to options
 *   flist - comma separated list of functions to break
 * Return: void
 */
void
xtr_set_func_breakpoints(struct opt *op)
{
	struct symbol *symp;
	char *ptrs[1024];		/* ptrs to function names */
	char *curr, *next, *flist;
	int i, set = 0;

	flist = strdup(op->flist);

	/*
	 * prepare ptrs. change each ',' to '\0' and put the pointer value
	 * into the array.
	 */
	for (curr = flist, i = 0; (next = strchr(curr, ',')) != NULL; i++) {
		ptrs[i] = curr;
		curr[next - curr] = '\0';
		curr = next + 1;
	}
	ptrs[i] = curr;
	ptrs[++i] = NULL;

	/* set bpt on every function if "-fall" flag given */
	if (!strcmp("all", op->flist)) {
		for (symp = NULL; (symp = xtr_get_next_sym()) != NULL; ) {
			xtr_set_bpt(op, symp);
			++set;
#ifdef __DEBUG__
			(void) fprintf(stderr, "*** bpt set on %s (%#x)\n",
			    symp->name, symp->va);
#endif /* __DEBUG__ */
		}
	}

	/* if not, loop over each function name */
	for (symp = NULL; (symp = xtr_get_next_sym()) != NULL; ) {
		for (i = 0; ptrs[i] != NULL; i++) {
			if (!strcmp(ptrs[i], symp->name)) {
				xtr_set_bpt(op, symp);
				++set;
#ifdef __DEBUG__
				(void) fprintf(stderr,
				    "*** bpt set on %s (%#x)\n",
				    symp->name, symp->va);
#endif /* __DEBUG__ */
			}
		}
	}

	free(flist);

	if (!set)
		(void) fprintf(stderr, "set_func_bpts: no breakpoints set\n");
}

/* 
 * Start the application, ordering to stop after execve()
 * Parameters:
 *    argv - arguments vector
 *    envp - environment
 * Return: void
 */
void
start_app(struct opt *op, char *const *argv, char *const *envp)
{
	op->flags |= XTR_STARTCHLD;
	op->exename = argv[0];
	op->pid = fork();

	/* if we're a child, set correct PID  */
	if (op->pid == (pid_t) 0)
		op->pid = getpid();
	else
		return;

	/* child stops */
	(void) kill(op->pid, SIGSTOP);

	/* set up the exec handler */
	xtr_before_exec(op);

	(void) execve(argv[0], argv, envp);

	/* if we're here, execve() failed */
}
