/*
 * Copyright (c) 2002 Toukan Ltd
 * All rights reserved
 *
 * Authors: Sergey Lyubka
 * Description:
 *   API definition for xtrace utility, see README for details.
 *   this file is a part of xtrace project
 *
 * List of platform-specific functions, that have to be defined for every
 * platform specific module:
 *
 *   o xtr_init
 *   o xtr_start_app
 *   o xtr_stop_app
 *   o xtr_resume_app
 *   o xtr_get_call_data
 *   o xtr_dump_mem
 *   o xtr_get_load_addr
 *   o xtr_get_sp
 *
 * $Id: xtrace.h,v 1.11 2002/05/15 14:40:57 drozd Exp $
 */

#ifndef __XTRACE__H
#define __XTRACE__H

#include <sys/types.h>

#include <stdlib.h>
#include <unistd.h>
#include <stddef.h>

typedef unsigned vaddr_t;

#define UNUSED(arg)	(arg) = (arg)

/*
 * Padding used to write syscalls return values
 */
#define PADDING			50

/* options */
struct opt {
	pid_t pid;		/* process's PID */
	unsigned flags;		/* see below for valid values */
	int scno;		/* syscall number */
	int signo;		/* signal number */
	int ncols;		/* characters printed */
	char *exename;		/* executable's full path name */
	char *flist;		/* comma-separated function list */
	char *slist;		/* comma-separated syscalls list */

	/* platform - specific stuff */
	void *data;			/* generic data */
	void (*fini)(struct opt *);	/* cleanup */
	void (*child)(int);		/* SIGCLD handler ??? */
};

#define XTR_BACKTRACE	(1 << 0)	/* dump per-thread backtrace */
#define XTR_TIMESTAMP	(1 << 1)	/* dump timestamp */
#define XTR_FUNCLIST	(1 << 2)	/* dump function list and exit */
#define XTR_DUMPMEM	(1 << 3)	/* dump memory content for parameters */
#define XTR_TRACESYS	(1 << 4)	/* trace system calls */
#define XTR_TRACEFUNC	(1 << 5)	/* trace functions */
#define XTR_DUMPTHREADS	(1 << 6)	/* dump per-thread info */
#define XTR_INSYSENTER	(1 << 7)	/* stopped in syscall entry */
#define XTR_INSYSEXIT	(1 << 8)	/* stopped in syscall exit */
#define XTR_INFENTER	(1 << 9)	/* stopped in fuction call enter */
#define XTR_INFEXIT	(1 << 10)	/* stopped in fuction exit */
#define XTR_INEXEC	(1 << 11)	/* stopped in execve() syscall */
#define XTR_SIGNAL	(1 << 12)	/* stopped because of a signal */
#define XTR_OTHER	(1 << 13)	/* stopped by some other reason */
#define XTR_CHILDEXITED	(1 << 20)	/* child has called exit() */
#define XTR_STARTCHLD	(1 << 21)	/* xtrace has called start_app */
#define XTR_INTERRUPTED	(1 << 22)	/* syscall was interrupted by signal */
#define XTR_INHANDLER	(1 << 23)	/* xtrace is in SIGCHLD  handler */
#define XTR_STEPPING	(1 << 24)	/* child single steps */
#define XTR_ATTACHED	(1 << 25)	/* attached to a child */
#define XTR_BACKTRACING	(1 << 26)	/* xtrace is doing backtrace */


#define CLEAR_FLAGS(op)						\
do {								\
	op->flags &= ~XTR_INFEXIT;				\
	op->flags &= ~XTR_INFENTER;				\
	op->flags &= ~XTR_INSYSEXIT;				\
	op->flags &= ~XTR_INSYSENTER;				\
	op->flags &= ~XTR_INEXEC;				\
	op->flags &= ~XTR_SIGNAL;				\
} while (0)


/* symbol table */
struct symbol {
	vaddr_t va;		/* virtual address in debuggee address space */
	char *name;		/* symbol's name */
	unsigned long instr;	/* instruction that lies on that address */
	unsigned flags;		/* see below for valid values */
};

#define SYM_REFERENCED		(1 << 0)	/* referenced from executable */
#define SYM_SETWATCH		(1 << 1)	/* watchpoint can be set */
#define SYM_BPT_SET		(1 << 2)	/* breakpoint is set */
#define SYM_BPT_REQUESTED	(1 << 3)	/* breakpoint requested */

/* In or Out parameter */
enum ArgFlow { IN, OUT };

/* parameter's type */
enum ArgType { HEX = 1, INT, STRING, SOCKADDR, MSGHDR, SIGALTSTACK,
	IOCTL, ARGV, ITIMER, FCNTL, TIMEVAL, TIMEZONE, IOVEC, STATFS,
	RTPRIO, STAT, RLIMIT, TIMESPEC, MODSTAT, KLDSTAT, SHEDPARAM,
	SIGSET, SIGACTION, KEVENT, FDSET };

struct Arg {
	enum ArgType type;
	enum ArgFlow inout;
};

#define MAXARGS		10
struct syscall {
	int no;					/* syscall's number */
	const char *name;			/* syscall's name */
	int nargs;				/* number of parameters */
	struct Arg args[MAXARGS];		/* MAXARGS parameters max */
};

/* function call data */
struct calldata {
	vaddr_t ip;		/* instruction pointer */
	vaddr_t retval;		/* return value */
	vaddr_t fp;		/* frame pointer */
	vaddr_t rta;		/* return address */
	vaddr_t ofp;		/* previous frame pointer */
	vaddr_t **argv;		/* array of arguments */
};

struct sharedlib {
	char *path;		/* full path to dependant library */
	vaddr_t base;		/* load address */
};

/* initialize tracer */
extern int xtr_init(struct opt *, char *, char *);

/* find symbol for given virtual address */
extern struct symbol *xtr_find_symbol(vaddr_t, ptrdiff_t *);

/* get list of dependant ELF binaries */
extern struct sharedlib **xtr_get_deps(pid_t);

/* build symbol table */
extern int xtr_symtab_build(const struct opt *);

/* destroy symbol table */
extern void xtr_symtab_destroy(void);

/* get syscall's character string by number */
extern const struct syscall *xtr_lookup_syscall(int);

/* stop on event of interest */
extern int xtr_stop_app(struct opt *);

/* resume process */
extern int xtr_resume_app(struct opt *);

/* iterate over symbolint xtr_stop_app(const struct opt *);s */
extern struct symbol *xtr_get_next_sym(void);

/* read memory content from process's address space */
extern int xtr_read_mem(const struct opt *, vaddr_t, void *, int);

/* write memory content into process's address space */
extern int xtr_write_mem(const struct opt *, vaddr_t, void *, int);

/* get call data by stack pointer */
extern int xtr_get_call_data(struct opt *, struct calldata *);

/* set callback to be called after forking a child */
extern void xtr_before_exec(struct opt *);

/* dump call */
extern void xtr_dump_call(struct opt *);

/* set breakpoint on function */
extern int xtr_set_bpt(struct opt *, struct symbol *);

/* clear breakpoint */
extern int xtr_clear_bpt(struct opt *, struct symbol *);

/* set breakpoints on functions */
extern void xtr_set_func_breakpoints(struct opt *op);

#endif /* __XTRACE__H */
